;ATTiny2313 with external 4.194304 MHz crystal
;Fuses: low = 0xDD, high = 0xDB, ext = 0xFF

;simple quartz clock for 6 "intelligent" displays (BCD input)
;(C) Arne Rossius 2008-08-28, http://www.ebps.de.vu/
;[based on Tiny24 simple quartz clock]

;PORTS:
;  PB0-6 = #CS for seperate displays (left..right)
;  PB6   = "Set Minutes" key
;  PB7   = "Set Houts" key
;  PD0-3 = BCD Data Bit 0..3 for all displays
;  PD4   = #WR for all displays
;
;VERSION HISTORY
;  2009-01-06  Arne Rossius
;    * added support for non-inverted latch enable inputs (e.g. 4543)

.equ	KEY_REPEATSTART = 16 ;16/16s = 1s until repeating starts
.equ	KEY_REPEAT = 5 ;16/5 = 3.2 repeats per second

.equ	InvertedLatch = 0 ;set to 1 for active-low latch enable inputs

.include "../include/tn2313def.inc"
 
.def	temp = R16
.def	temp2 = R17
.def	sec1 = R18
.def	sec10 = R19
.def	min1 = R20
.def	min10 = R21
.def	hr1 = R22
.def	hr10 = R23
.def	disp = R24
.def	repeat = R25

;R31:R30: Z pointer for Flash access

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org OVF0addr
	rjmp	keyrepeat
.org PCINTaddr
	rjmp	keytoggle
	
;===============================================================================

.if (InvertedLatch == 0)
.macro Latch1
sbi @0, @1
.endmacro
.macro Latch0
cbi @0, @1
.endmacro
.else
.macro Latch1
cbi @0, @1
.endmacro
.macro Latch0
sbi @0, @1
.endmacro
.endif

;===============================================================================

keytoggle:
	;pin change interrupt for hour/minute keys
	push	temp
	in	temp, SREG
	push	temp
	
	in	temp, PINB
	andi	temp, 0xC0
	cpi	temp, 0xC0
	breq	key_released
	
	;key pressed: reset and enable debounce/repeat timer
	ldi	temp, 128
	out	TCNT0, temp
	ldi	temp, 0x04 ;Clk/256 => ~8ms debouncing time
	out	TCCR0B, temp
	
keytoggle_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_released:
	;key released: disable debounce/repeat timer
	ldi	temp, 0x00
	out	TCCR0B, temp
	ldi	repeat, 0
	rjmp	keytoggle_end

;===============================================================================

keyrepeat:
	;key debounce/repeat timer overflow
	push	temp
	in	temp, SREG
	push	temp
	
	;change from debounce to repeat mode
	cpi	repeat, 0
	brne	repeatdiv
	ldi	temp, 0x05 ;Clk/1024
	out	TCCR0B, temp
	ldi	temp, 0
	out	TCNT0, temp
	ldi	repeat, KEY_REPEATSTART
	rjmp	key_action
	
repeatdiv:
	;repeated keypress
	dec	repeat
	brne	keyrepeat_end
	ldi	repeat, KEY_REPEAT
key_action:
	;perform key action
	sbis	PINB, 7
	rjmp	key_hr
	sbis	PINB, 6
	rjmp	key_min
	
keyrepeat_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_hr:
	;"Hour" key: increment hours
	rcall	inc_hr
	rcall	update_disp
	rjmp	keyrepeat_end

key_min:
	;"Minute" key: increment minutes and reset seconds
	rcall	inc_min
	clr	sec10
	clr	sec1
	rcall	update_disp
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	rjmp	keyrepeat_end
	
;===============================================================================

reset:
	;stack pointer setup
	ldi	temp, RAMEND
	out	SPL, temp
	
	;port setup
	ldi	temp, 0x3F
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	ldi	temp, 0x70
	out	PORTD, temp
	
	;Timer 0 setup (Key debounce/repeat timer)
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;Timer 1 setup (Clock timer)
	ldi	temp, 1<<WGM12 | 0x05 ;CTC Mode, Clk/1024
	out	TCCR1B, temp
	ldi	temp, HIGH(4096) ;overflow every second (4194304 Hz / (1024 * 4096) = 1 Hz)
	out	OCR1AH, temp
	ldi	temp, LOW(4096)
	out	OCR1AL, temp
	
	;Pin Change interrupt setup
	ldi	temp, 0xC0
	out	PCMSK, temp
	ldi	temp, 1<<PCIE
	out	GIMSK, temp
	
	;register setup
	ldi	hr10, 0
	ldi	hr1, 0
	ldi	min10, 0
	ldi	min1, 0
	ldi	sec10, 0
	ldi	sec1, 0
	rcall	update_disp
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------
	
loop:
	;increment clock if timer reached 1s
	in	temp, TIFR
	sbrs	temp, OCF1A
	rjmp	timeinc_end
	ldi	temp, 1<<OCF1A ;reset flag
	out	TIFR, temp
	rcall	inc_sec
	rcall	update_disp
timeinc_end:
	rjmp	loop
	
;===============================================================================

inc_sec:
	;increment seconds
	inc	sec1
	cpi	sec1, 10
	brlo	inc_end
	clr	sec1
	inc	sec10
	cpi	sec10, 6
	brlo	inc_end
	clr	sec10
inc_min:
	;increment minutes
	inc	min1
	cpi	min1, 10
	brlo	inc_end
	clr	min1
	inc	min10
	cpi	min10, 6
	brlo	inc_end
	clr	min10
inc_hr:
	;increment hours
	inc	hr1
	cpi	hr1, 10
	brlo	hr_24
	clr	hr1
	inc	hr10
	rjmp	inc_end
hr_24:
	cpi	hr1, 4
	brne	inc_end
	cpi	hr10, 2
	brne	inc_end
	clr	hr1
	clr	hr10
inc_end:
	ret
	
;===============================================================================

update_disp:
	;display new time
	mov	temp, hr10
	ori	temp, 0x70
	out	PORTD, temp
	rjmp	PC+1
	Latch1	PORTB, 0
	rjmp	PC+1
	cbi	PORTD, 4
	rjmp	PC+1
	sbi	PORTD, 4
	rjmp	PC+1
	Latch0	PORTB, 0
	
	mov	temp, hr1
	ori	temp, 0x70
	out	PORTD, temp
	rjmp	PC+1
	Latch1	PORTB, 1
	rjmp	PC+1
	cbi	PORTD, 4
	rjmp	PC+1
	sbi	PORTD, 4
	rjmp	PC+1
	Latch0	PORTB, 1

	mov	temp, min10
	ori	temp, 0x70
	out	PORTD, temp
	rjmp	PC+1
	Latch1	PORTB, 2
	rjmp	PC+1
	cbi	PORTD, 4
	rjmp	PC+1
	sbi	PORTD, 4
	rjmp	PC+1
	Latch0	PORTB, 2

	mov	temp, min1
	ori	temp, 0x70
	out	PORTD, temp
	rjmp	PC+1
	Latch1	PORTB, 3
	rjmp	PC+1
	cbi	PORTD, 4
	rjmp	PC+1
	sbi	PORTD, 4
	rjmp	PC+1
	Latch0	PORTB, 3

	mov	temp, sec10
	ori	temp, 0x70
	out	PORTD, temp
	rjmp	PC+1
	Latch1	PORTB, 4
	rjmp	PC+1
	cbi	PORTD, 4
	rjmp	PC+1
	sbi	PORTD, 4
	rjmp	PC+1
	Latch0	PORTB, 4

	mov	temp, sec1
	ori	temp, 0x70
	out	PORTD, temp
	rjmp	PC+1
	Latch1	PORTB, 5
	rjmp	PC+1
	cbi	PORTD, 4
	rjmp	PC+1
	sbi	PORTD, 4
	rjmp	PC+1
	Latch0	PORTB, 5

	ret
